/*
   init.c -- PIKT initialization routines

   $Header: /home/cvs/ucgsb/pikt/src/piktd/init.c,v 1.19 1999/06/21 15:39:47 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#if defined(SOLARIS)
	#include <sys/systeminfo.h>
#endif

#include "common.h"

int ismaster;

char *home,
     *bindir,
     *etcdir,
     *libdir,
     *vardir,
     *altdir,
     *prgdir,
     *objdir,
     *cfgdir,
     *difdir,
     *stgdir,
     *hstdir,
     *logdir,
     *PIKTconf,
     *pikt,
     *piktd,
     *piktc,
     *piktc_svc,
     *piktdconf,
     *piktdconftmp,
     *piktdlock,
     *piktdlog,
     *piktdalt,
     *piktclock,
     *piktclog,
     *piktc_svclock,
     *piktc_svclog;

int isopt (char **, char **, int);
void gethome(char **);
void initpaths(char *);
void setservice(unsigned long, char *);
void *initmalloc(size_t, char *);

/* perform various initializations
 */
void
init (int argc, char **argv, void (*usage)(void))
{
	char *cp;
	char *fqhn;

	istest = isopt(argv+1, argv+argc-1, 'T');
	isdebug = isopt(argv+1, argv+argc-1, 'G');
	if((ishelp = isopt(argv+1, argv+argc-1, 'h'))) {
		(*usage)();
		exit(1);
	}

	gethome(argv);		/* determine PIKT home dir */

	initpaths(home);	/* initialize dir/file paths */

	/* from here on, it is safe to call either logreturn() or logexit()
	   since we have the paths to the log files
	 */

	gethostname(hostname, MAXIDENT);
	cp = strchr(hostname, '.');
	if (cp) {
		fqhn = strdup(hostname);
		if (! istest)
			*cp = '\0';
		strcpy(domainname, cp+1);
	} else {
		fqhn = (char *)NULL;
		/* what we really want here is not the nis domain,
		   but rather the internet domainname
		 */
		#ifdef SOLARIS
		sysinfo(SI_SRPC_DOMAIN, domainname, MAXIDENT);
		#else
		getdomainname(domainname, MAXIDENT);
		#endif
	}
	if (istest) {
		master = strdup(hostname);
		domain = strdup("domain");
		accessmode = strdup("local");
	} else
		getconf();	/* get config stuff from PIKT.conf */

	if (fqhn) {
		if (!strcmp(fqhn, master))
			ismaster = TRUE;
		free(fqhn);
	} else {
		if (!strncmp(hostname, master, (size_t)strlen(hostname)))
			ismaster = TRUE;
	}

	sig_init();		/* initialize signal handlers */

	return;
}

/* look for a command-line (minus) switch, returning TRUE if found, else
   FALSE; this is for simpler command-line parsing; for more involved parsing,
   we use cmd.[lyc] and cmdparse()
 */
int
isopt (char **avcurr, char **avlast, int opt)
{
	while (avcurr <= avlast) {
		if (**avcurr == '-') {
			if (strchr(*avcurr, opt))
				return TRUE;
		}
		avcurr++;
	}
	return FALSE;
}

/* determine the PIKT home directory, either directly from argv (when
   the program is invoked with a full path), or by searching through the
   environment PATH (when the program is invoked with no path)
 */
void
gethome (char **progpath)
{
	char *pp;
	static char *envpath, *path, *prog, *pathprog;
	int proglen, homelen;
	struct stat fbuf;
	char dircur[MAXPATH+1];
	int found;

	/* note that for error logging, we can't invoke logreturn() or
	   logexit(), because we don't have the path to the log file yet!
	 */

	home = prog = strdup(*progpath);
	proglen = strlen(prog);
	if ((pp = strrchr(home, '/'))) {
		*pp = '\0';
		if (!strcmp(home, ".")) {	/* prog invoked as ./<prog> */
			if (getcwd(dircur, MAXPATH) == NULL) {
				fprintf(stderr, "in gethome(), getcwd() failure, cannot determine current working directory\n");
				fflush(stderr);
				exit(1);
			}
			free(home);
			home = strdup(dircur);
		}
	} else {	/* there is no explicit path, so we have to get it from
	           the PATH environment variable
	         */
		if (! (pp = getenv("PATH"))) {
			fprintf(stderr, "in gethome(), cannot determine environment PATH\n");
			fflush(stderr);
			exit(1);
		} else
			envpath = strdup(pp);
		/* traverse the PATH, until we find PIKT home */
		if (strtok(envpath, ": ") == NULL) {
			fprintf(stderr, "in gethome(), cannot tokenize environment PATH\n");
			fflush(stderr);
			exit(1);
		} else
			path = envpath;
		do {
			pathprog = (char *) initmalloc((strlen(path)+2+proglen) * sizeof(char), "gethome");
			sprintf(pathprog, "%s/%s", path, prog);
			found = ! lstat(pathprog, &fbuf);
			free(pathprog);
			if (found) {
				free(prog);
				home = strdup(path);
				break;
			}
		} while ((path = strtok(NULL, ": ")));
		if (! path) {
			fprintf(stderr, "in gethome(), %s not in environment PATH\n", prog);
			fflush(stderr);
			exit(1);
		}
		free(envpath);
	}
	homelen = strlen(home);
	if ((homelen >= 4) && ! strcmp(home+homelen-4, "/bin"))
		*(home+homelen-4) = '\0';
	else {
		fprintf(stderr, "in gethome(), cannot find PIKT bin dir\n");
		fflush(stderr);
		exit(1);
	}
	return;
}

/* get PIKT.conf file settings; in the future, as we add stuff to PIKT.conf,
   we will have to rewrite this in a more general way, else we'll need to
   develop conf.[ly] and confparse() ala cfg.[ly] and cfgparse()
 */
void
getconf (void)
{
	FILE *conffile;
	char *cp;
	char line[MAXLINE+1], parm[MAXLINE+1], value[MAXLINE+1];

	if ((conffile = fopen(PIKTconf, "r")) == NULL)
		logexit(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in getconf(), missing %s file", PIKTconf);
	while (fgets(line, MAXLINE+1, conffile) != NULL) {
		/* skip blank lines */
		if (line[0] == '\n')
			continue;
		/* disregard comments */
		if (    (cp = strstr(line, "//"))
		     || (cp = strstr(line, "#"))
		   )
			*cp = '\0';
		if (sscanf(line, "%s %s", parm, value) == 2) {

			/* mandatory access parameters */

			if (! strcmp(parm, "uid")) {
				errno = 0;
				uid = atoi(value);
				if (errno)
					logexit(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in %s, invalid uid %s", PIKTconf, value);
			} else if (! strcmp(parm, "master"))
				master = strdup(value);
			else if (! strcmp(parm, "domain"))
				domain = strdup(value);
			else if (! strcmp(parm, "access"))
				accessmode = strdup(value);

			/* optional access parameters */

			/* all services */
			else if (! strcmp(parm, "all_services"))
				setservice(ALL_SERVICES, value);

			/* kill services */
			else if (! strcmp(parm, "kill_piktd_service"))
				setservice(KILL_PIKTD_SERVICE, value);
			else if (! strcmp(parm, "kill_piktc_svc_service"))
				setservice(KILL_PIKTC_SVC_SERVICE, value);
			else if (! strcmp(parm, "kill_service"))
				setservice(KILL_SERVICE, value);

			/* restart services */
			else if (! strcmp(parm, "restart_piktd_service"))
				setservice(RESTART_PIKTD_SERVICE, value);
			else if (! strcmp(parm, "restart_piktc_svc_service"))
				setservice(RESTART_PIKTC_SVC_SERVICE, value);
			else if (! strcmp(parm, "restart_service"))
				setservice(RESTART_SERVICE, value);

			/* call_back; note that call_back_service is
			   always authorized
			 */
			else if (! strcmp(parm, "call_back"))
				setservice(CALL_BACK_SERVICE, value);

			/* alerts services */
			else if (! strcmp(parm, "disable_alerts_service"))
				setservice(DISABLE_ALERTS_SERVICE, value);
			else if (! strcmp(parm, "enable_alerts_service"))
				setservice(ENABLE_ALERTS_SERVICE, value);
			else if (! strcmp(parm, "execute_alerts_service"))
				setservice(EXECUTE_ALERTS_SERVICE, value);
			else if (! strcmp(parm, "execute_scripts_service"))
				setservice(EXECUTE_ALERTS_SERVICE, value);
			else if (! strcmp(parm, "show_alerts_status_service"))
				setservice(SHOW_ALERTS_STATUS_SERVICE, value);

			/* install services */
			else if (! strcmp(parm, "install_alerts_service"))
				setservice(INSTALL_ALERTS_SERVICE, value);
			else if (! strcmp(parm, "install_scripts_service"))
				setservice(INSTALL_ALERTS_SERVICE, value);
			else if (! strcmp(parm, "install_programs_service"))
				setservice(INSTALL_PROGRAMS_SERVICE, value);
			else if (! strcmp(parm, "install_files_service"))
				setservice(INSTALL_FILES_SERVICE, value);
			else if (! strcmp(parm, "install_objects_service"))
				setservice(INSTALL_OBJECTS_SERVICE, value);
			else if (! strcmp(parm, "install_service"))
				setservice(INSTALL_SERVICE, value);

			/* fetch services */
			else if (! strcmp(parm, "fetch_alerts_service"))
				setservice(FETCH_ALERTS_SERVICE, value);
			else if (! strcmp(parm, "fetch_scripts_service"))
				setservice(FETCH_ALERTS_SERVICE, value);
			else if (! strcmp(parm, "fetch_programs_service"))
				setservice(FETCH_PROGRAMS_SERVICE, value);
			else if (! strcmp(parm, "fetch_files_service"))
				setservice(FETCH_FILES_SERVICE, value);
			else if (! strcmp(parm, "fetch_objects_service"))
				setservice(FETCH_OBJECTS_SERVICE, value);
			else if (! strcmp(parm, "fetch_service"))
				setservice(FETCH_SERVICE, value);

			/* delete services */
			else if (! strcmp(parm, "delete_alerts_service"))
				setservice(DELETE_ALERTS_SERVICE, value);
			else if (! strcmp(parm, "delete_scripts_service"))
				setservice(DELETE_ALERTS_SERVICE, value);
			else if (! strcmp(parm, "delete_programs_service"))
				setservice(DELETE_PROGRAMS_SERVICE, value);
			else if (! strcmp(parm, "delete_files_service"))
				setservice(DELETE_FILES_SERVICE, value);
			else if (! strcmp(parm, "delete_objects_service"))
				setservice(DELETE_OBJECTS_SERVICE, value);
			else if (! strcmp(parm, "delete_service"))
				setservice(DELETE_SERVICE, value);

			/* list services */
			else if (! strcmp(parm, "list_alerts_service"))
				setservice(LIST_ALERTS_SERVICE, value);
			else if (! strcmp(parm, "list_scripts_service"))
				setservice(LIST_ALERTS_SERVICE, value);
			else if (! strcmp(parm, "list_programs_service"))
				setservice(LIST_PROGRAMS_SERVICE, value);
			else if (! strcmp(parm, "list_files_service"))
				setservice(LIST_FILES_SERVICE, value);
			else if (! strcmp(parm, "list_objects_service"))
				setservice(LIST_OBJECTS_SERVICE, value);
			else if (! strcmp(parm, "list_service"))
				setservice(LIST_SERVICE, value);

			/* other parameters */

			else if (! strcmp(parm, "rpc_timeout")) {
				errno = 0;
				rpc_timeout = atol(value);
				if (errno)
					logexit(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in %s, invalid rpc_timeout %s", PIKTconf, value);

			} else if (! strcmp(parm, "rpc_retry_timeout")) {
				errno = 0;
				rpc_retry_timeout = atol(value);
				if (errno)
					logexit(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in %s, invalid rpc_retry_timeout %s", PIKTconf, value);

			} else
				logexit(PIKTLOG, ERR, __FILE__, __LINE__, FALSE, "in %s, unknown parameter %s", PIKTconf, parm);

		}
	}
	fclose(conffile);
	return;	
}

/* initialize dir/file paths
 */
void
initpaths (char *hom)
{
	const char *testbranch;

	if (istest)
		testbranch = "/src/test";
	else
		testbranch = "";

	/* there are lots of redundant calls to strlen() here; we could
	   save strlen() return values to local vars, but in the interest
	   of simplicity we choose not to
	 */

	bindir = (char *) initmalloc((strlen(hom)+5) * sizeof(char), "initpaths");
	sprintf(bindir, "%s/bin", hom);		/* we don't insert testbranch
						   into bindir
						 */

	etcdir = (char *) initmalloc((strlen(hom)+strlen(testbranch)+5) * sizeof(char), "initpaths");
	sprintf(etcdir, "%s%s/etc", hom, testbranch);

	libdir = (char *) initmalloc((strlen(hom)+strlen(testbranch)+5) * sizeof(char), "initpaths");
	sprintf(libdir, "%s%s/lib", hom, testbranch);

	vardir = (char *) initmalloc((strlen(hom)+strlen(testbranch)+5) * sizeof(char), "initpaths");
	sprintf(vardir, "%s%s/var", hom, testbranch);

	altdir = (char *) initmalloc((strlen(libdir)+8) * sizeof(char), "initpaths");
	sprintf(altdir, "%s/alerts", libdir);

	prgdir = (char *) initmalloc((strlen(libdir)+10) * sizeof(char), "initpaths");
	sprintf(prgdir, "%s/programs", libdir);

	objdir = (char *) initmalloc((strlen(libdir)+9) * sizeof(char), "initpaths");
	sprintf(objdir, "%s/objects", libdir);

	cfgdir = (char *) initmalloc((strlen(libdir)+9) * sizeof(char), "initpaths");
	sprintf(cfgdir, "%s/configs", libdir);

	difdir = (char *) initmalloc((strlen(cfgdir)+9) * sizeof(char), "initpaths");
	sprintf(difdir, "%s/diffing", cfgdir);

	stgdir = (char *) initmalloc((strlen(cfgdir)+9) * sizeof(char), "initpaths");
	sprintf(stgdir, "%s/staging", cfgdir);

	hstdir = (char *) initmalloc((strlen(vardir)+11) * sizeof(char), "initpaths");
	sprintf(hstdir, "%s/histories", vardir);

	logdir = (char *) initmalloc((strlen(vardir)+5) * sizeof(char), "initpaths");
	sprintf(logdir, "%s/log", vardir);

	PIKTconf = (char *) initmalloc((strlen(etcdir)+11) * sizeof(char), "initpaths");
	sprintf(PIKTconf, "%s/PIKT.conf", etcdir);

	pikt = (char *) initmalloc((strlen(bindir)+6) * sizeof(char), "initpaths");
	sprintf(pikt, "%s/pikt", bindir);

	piktd = (char *) initmalloc((strlen(bindir)+7) * sizeof(char), "initpaths");
	sprintf(piktd, "%s/piktd", bindir);

	piktc = (char *) initmalloc((strlen(bindir)+7) * sizeof(char), "initpaths");
	sprintf(piktc, "%s/piktc", bindir);

	piktc_svc = (char *) initmalloc((strlen(bindir)+11) * sizeof(char), "initpaths");
	sprintf(piktc_svc, "%s/piktc_svc", bindir);

	piktdconf = (char *) initmalloc((strlen(etcdir)+12) * sizeof(char), "initpaths");
	sprintf(piktdconf, "%s/piktd.conf", etcdir);

	piktdconftmp = (char *) initmalloc((strlen(piktdconf)+5) * sizeof(char), "initpaths");
	sprintf(piktdconftmp, "%s.tmp", piktdconf);

	piktdlock = (char *) initmalloc((strlen(etcdir)+12) * sizeof(char), "initpaths");
	sprintf(piktdlock, "%s/piktd.lock", etcdir);

	piktdlog = (char *) initmalloc((strlen(logdir)+11) * sizeof(char), "initpaths");
	sprintf(piktdlog, "%s/piktd.log", logdir);

	piktdalt = (char *) initmalloc((strlen(etcdir)+8) * sizeof(char), "initpaths");
	sprintf(piktdalt, "%s/alerts", etcdir);

	piktclock = (char *) initmalloc((strlen(etcdir)+12) * sizeof(char), "initpaths");
	sprintf(piktclock, "%s/piktc.lock", etcdir);

	piktclog = (char *) initmalloc((strlen(logdir)+11) * sizeof(char), "initpaths");
	sprintf(piktclog, "%s/piktc.log", logdir);

	piktc_svclock = (char *) initmalloc((strlen(etcdir)+16) * sizeof(char), "initpaths");
	sprintf(piktc_svclock, "%s/piktc_svc.lock", etcdir);

	piktc_svclog = (char *) initmalloc((strlen(logdir)+15) * sizeof(char), "initpaths");
	sprintf(piktc_svclog, "%s/piktc_svc.log", logdir);

	return;
}

/* set/unset the specified bit in services, thereby enabling/disabling the
   specified piktc_svc service
 */
void
setservice (unsigned long service, char *setting)
{

	if (!strcasecmp("yes", setting) || !strcasecmp("true", setting) || !strcasecmp("on", setting)) {
		services |= service;
	} else if (!strcasecmp("no", setting) || !strcasecmp("false", setting) || !strcasecmp("off", setting)) {
		services &= ~service;
	} else {
		logexit(PIKTLOG, ERR, __FILE__, __LINE__, FALSE, "in setservices(), invalid service setting");
	}
	return;
}

/* do the first and second given host names match?
   'vienna' matches 'vienna'
   'vienna' matches 'vienna.uppity.edu'
   'vienna.uppity.edu' matches 'vienna'
   'acme.com' matches 'acme.com.uk' and vice-versa--practically speaking,
   is this a problem?  we could check for this weird case, but at the cost
   of slowing down all match checks
 */
int
hostmatch (char *hn1, char *hn2)
{
	register int i;
	register char c1, c2;

	i = 0;
	do {
		c1 = *(hn1+i);
		c2 = *(hn2+i);
		if (!c1 || !c2)
			break;
		if (c1 != c2)
			return FALSE;
		i++;
	} while (TRUE);
	if (!c1) {
		if (!c2 || (c2 == '.'))
			return TRUE;
	} else if (!c2) {
		if (!c1 || (c1 == '.'))
			return TRUE;
	}
	return FALSE;
}

/* malloc() the specified amount of memory; if unsuccessful, presumably
   because of virtual memory exhaustion, output a warning to stderr and
   keep trying; if successful, return a generic pointer to the newly
   allocated memory; we can't use mymalloc(), because it calls logreturn(),
   and we don't have the path to the logdir yet
 */
void *
initmalloc (size_t size, char *func)
{
	void *ptr;
	static char msg[64];	/* should be plenty */

	while (! (ptr = malloc(size))) {
		sprintf(msg, "in %s(), malloc() failure", func);
		fprintf(stderr, msg);
		fflush(stderr);
		sleep(MALLOC_TIME_TO_SLEEP);
	}
	return ptr;
}

/* return TRUE if file exists, FALSE if not; only call after initpaths()
   so that we may invoke logreturn()
 */
int
isfile (char *file)
{
	struct stat fbuf;

	if (lstat(file, &fbuf)) {
		logreturn(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in isfile(), stat() error for %s", file);
		return FALSE;
	} else
		return TRUE;
}
